import os
import re
import sys
import time
import shutil
import logging

import boto

from easyaws.libs import interface, aws, ssh


logger = logging.getLogger(__name__)


class FindDjangoProject():
    __ignore_patterns = (
        # Bytecode
        '__pycache__', '.pyc',

        # Git
        '.git', '.gitignore',

        # Svn
        '.svn',

        # Cvs
        'CVS',

        # PyCharm
        '.idea',
    )

    def __init__(self):
        """
        Find django project in the current directory
        """

        self.project_path = os.getcwd()
        self.manage_path = os.path.join(self.project_path, 'manage.py')

        logger.debug('Try to find django project on the path: {}'.format(self.project_path))

        # Check manage.py
        if not os.path.isfile(self.manage_path):
            print('File manage.py is not found')
            sys.exit(1)
        self.project_path = os.getcwd()
        logger.debug('File manage.py is found')

        # Find name of the django project
        with open(self.manage_path, 'r') as manage:
            pattern = r'"(\w+)\.settings"'
            result = re.search(pattern, manage.read())
            if not result:
                print('Name of the django project is not found in manage.py')
                sys.exit(1)
        self.project_name = result.group(1)
        logger.debug('Name of the project is "{}"'.format(self.project_name))

        # Check wsgi.py
        self.wsgi_path = os.path.join(self.project_path, self.project_name, 'wsgi.py')
        if not os.path.isfile(self.wsgi_path):
            print('File wsgi.py is not found in the project directory')
            sys.exit(1)
        logger.debug('File manage.py is found')

        # Generate id of the project contains letter or digits only for using in aws
        self.aws_name = ''.join([i for i in self.project_name if i.isalnum()])
        logger.debug('AWS name of the project is "{}"'.format(self.aws_name))

    def create_archive(self, archive_format='gztar'):
        """
        Archive cleaned django project (without git, bytocide etc)

        :type archive_format: str
        :param archive_format: format of archive

        :return: path to archive
        """

        # Check archive format
        if archive_format not in [a[0] for a in shutil.get_archive_formats()]:
            logger.critical('Unsupported archive format: {}'.format(archive_format))
            sys.exit(1)

        # Copy project to temp directory
        logger.debug('Copy django project to temporary directory')
        temp_path = '/tmp/{}'.format(self.project_name)
        shutil.rmtree(temp_path, ignore_errors=True)
        shutil.copytree(self.project_path, temp_path, ignore=shutil.ignore_patterns(*self.__ignore_patterns))

        # Make archive
        logger.debug('Archive django project')
        path = shutil.make_archive('/tmp/{}'.format(self.project_name), archive_format, temp_path)

        # Return archive path
        return path


class Commands:
    @classmethod
    def configure(cls):
        """
        Configure boto, i.e enter AWS credentials and select region
        """

        # Request credentials
        aws_access_key_id = interface.request(
            'Enter AWS_ACCESS_KEY_ID',
            boto.config.get('Credentials', 'aws_access_key_id'))
        aws_secret_access_key = interface.request(
            'Enter AWS_SECRET_ACCESS_KEY',
            boto.config.get('Credentials', 'aws_secret_access_key'))

        # Request region
        aws_region_name = interface.select('Select region', aws.get_regions())

        # Save them to file
        config_path = os.path.join(os.path.expanduser('~'), '.boto')
        with open(config_path, 'w') as f:
            content = str(
                '[Credentials]\n'
                'aws_access_key_id = {aws_access_key_id}\n'
                'aws_secret_access_key = {aws_secret_access_key}\n'
                '\n'
                '[Boto]\n'
                'aws_region_name = {aws_region_name}'  # for libs.aws module
            ).format(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_region_name=aws_region_name
            )
            f.write(content)

        # Hide config file
        os.chmod(config_path, 0o600)

    @classmethod
    def clean(cls):
        """
        Full clean AWS cloud, i.e remove all auto scaling groups, launch configurations, load balancers, images,
        instances, key pairs and security groups
        """

        # Request confirmation
        if not interface.confirm():
            return

        # Remove all auto scaling groups
        for group in aws.connections.autoscale.get_all_groups():
            aws.helpers.delete_auto_scaling_group(group.name)

        # Remove all launch configurations
        for config in aws.connections.autoscale.get_all_launch_configurations():
            aws.helpers.delete_launch_configuration(config.name)

        # Remove all load balancers
        for balancer in aws.connections.elb.get_all_load_balancers():
            aws.helpers.delete_load_balancer(balancer.name)

        # Remove all images
        for image in aws.helpers.get_all_images():
            aws.helpers.delete_image(image)

        # Remove all instances
        for instance in aws.helpers.get_all_instances():
            aws.helpers.delete_instance(instance)

        # Remove all key pairs
        for key in aws.connections.ec2.get_all_key_pairs():
            aws.helpers.delete_key_pair(key.name)

        # Remove all security groups
        for security_group in aws.connections.ec2.get_all_security_groups():
            if security_group.name == 'default':
                continue
            aws.helpers.delete_security_group(security_group.name)

    class Project:
        """
        Deploy django project
        """

        @classmethod
        def push(cls):
            """
            Create and setup temporary instance, upload your django project on instance, create image and remove
            instance
            """

            # Find django project
            django_project = FindDjangoProject()

            # Create security group
            security_group = aws.helpers.get_or_create_security_group('develop', [22, 80])

            # Create key pair
            key_pair = aws.helpers.get_or_create_key_pair('develop')

            # Create unnamed instance
            instance = aws.helpers.create_instance(
                key_pair.name,
                security_group.name,
                aws.helpers.find_last_ubuntu_image_id('14.04'))

            # Prepare
            replacements = {
                'PROJECT_NAME': django_project.project_name,
                'PROJECT_PATH': '/home/ubuntu/{}'.format(django_project.project_name),
                'VIRTUALENV_PATH': '/home/ubuntu/virtualenv',
            }
            packages = ['python-pip']
            files = [
                (django_project.create_archive(), '/tmp/project.tar.gz'),
                ('templates/nginx.conf', '/tmp/nginx.conf'),
                ('templates/nginx-app-proxy', '/tmp/nginx-app-proxy'),
                ('templates/gunicorn.conf.py', '/tmp/gunicorn.conf.py'),
                ('templates/supervisord.conf', '/tmp/my.supervisord.conf'),
                ('templates/supervisord-init', '/tmp/supervisord-init'),
            ]
            receipt = [
                # Create virtualenv
                'sudo pip install virtualenv',
                'virtualenv -p /usr/bin/python3 /home/ubuntu/virtualenv',

                # Extract django-project
                'mkdir {PROJECT_PATH}',
                'tar -xvzf /tmp/project.tar.gz -C {PROJECT_PATH}',

                # Install apt requirements
                '[ -f {PROJECT_PATH}/requirements-apt.txt ]'
                ' && xargs sudo apt-get install -y < {PROJECT_PATH}/requirements-apt.txt'
                ' || echo skip',

                # Install pip requirements
                '[ -f {PROJECT_PATH}/requirements.txt ]'
                ' && source /home/ubuntu/virtualenv/bin/activate'
                ' && pip install -r {PROJECT_PATH}/requirements.txt'
                ' || echo skip',

                # Install django in the virtualenv
                'source /home/ubuntu/virtualenv/bin/activate'
                ' && pip install django',

                # Install proxy-server Nginx and configure it
                'sudo apt-get install nginx -y',
                'sudo mv /etc/nginx/nginx.conf /etc/nginx/nginx.conf.old',
                'sudo mv /tmp/nginx.conf /etc/nginx/nginx.conf',
                'sudo chown root:root /etc/nginx/nginx.conf',
                'sudo rm -rf /etc/nginx/sites-enabled/default',
                'sudo mv /tmp/nginx-app-proxy /etc/nginx/sites-available/{PROJECT_NAME}',
                'sudo ln -sf /etc/nginx/sites-available/{PROJECT_NAME} /etc/nginx/sites-enabled/{PROJECT_NAME}',
                'sudo chown root:root /etc/nginx/sites-available/{PROJECT_NAME}',
                'sudo /etc/init.d/nginx restart',

                # Install Gunicorn in the virtualenv and configure it
                'source /home/ubuntu/virtualenv/bin/activate'
                ' && pip install gunicorn',
                'mv /tmp/gunicorn.conf.py {PROJECT_PATH}/gunicorn.conf.py',

                # Install Supervisor and configure it
                'sudo pip install supervisor',
                'echo_supervisord_conf > /tmp/supervisord.conf',
                'cat /tmp/my.supervisord.conf >> /tmp/supervisord.conf',
                'sudo mv /tmp/supervisord.conf /etc/supervisord.conf',
                'sudo supervisord',
                'sudo mv /tmp/supervisord-init /etc/init.d/supervisord',
                'sudo chmod +x /etc/init.d/supervisord',
                'sudo update-rc.d supervisord defaults',
            ]

            try:
                # Connect to the instance
                key_path = aws.helpers.generate_path_to_key_pair(key_pair.name)
                client = ssh.Client(instance.public_dns_name, key_path=key_path, replacements=replacements)

                # Setup instance
                client.put_files(files)
                client.apt_install(packages)
                client.run_commands(receipt)

                # Check availability of the instance
                aws.helpers.wait_host(instance.public_dns_name)

                # Create image based on the instance
                image_name = '{project_name}-{timestamp}'.format(
                    project_name=django_project.aws_name,
                    timestamp=time.strftime('%Y.%m.%d-%H.%M'))
                aws.helpers.create_image(image_name, instance)

            finally:
                # Remove instance
                aws.helpers.delete_instance(instance, wait=False)

        @classmethod
        def deploy(cls):
            """
            Create load balancer, create launch configuration based on the created image and create auto scaling group
            (auto scaling up to 3 instances when CPU-load > 70)
            """

            # Find django project
            django_project = FindDjangoProject()

            # Select image
            images = aws.helpers.get_all_images(filters={'name': django_project.aws_name + '*'}, reverse=True)
            image = interface.select('Please select image', images, title=lambda img: img.name or img.id)
            if not image:
                print('There are no images in the your cloud')
                sys.exit(1)

            # Create key
            key_pair = aws.helpers.get_or_create_key_pair(django_project.aws_name)

            # Create security group
            security_group = aws.helpers.get_or_create_security_group(django_project.aws_name)

            # Create new load balancer or get already created
            load_balancer = aws.helpers.create_load_balancer(
                django_project.aws_name,
                security_group.name,
                listeners=[(80, 80, 'http'), (443, 443, 'tcp')],
            )

            # Remove previous auto scaling group
            aws.helpers.delete_auto_scaling_group(django_project.aws_name)

            # Remove previous launch configuration
            aws.helpers.delete_launch_configuration(django_project.aws_name)

            # Create launch configuration
            launch_configuration = aws.helpers.create_launch_configuration(
                django_project.aws_name,
                image.id,
                key_pair.name,
                security_group.name,
            )

            # Create auto scaling group
            aws.helpers.create_auto_scaling_group(
                django_project.aws_name,
                load_balancer.name,
                launch_configuration.name,
            )

            # Check availability of the load balancer
            aws.helpers.wait_host(load_balancer.dns_name)

        @classmethod
        def undeploy(cls):
            """
            Remove auto scaling group and launch configuration
            """

            # Find django project
            django_project = FindDjangoProject()

            # Remove auto scaling group
            aws.helpers.delete_auto_scaling_group(django_project.aws_name)

            # Remove launch configuration
            aws.helpers.delete_launch_configuration(django_project.aws_name)

    class Renv:
        """
        Remote environment management
        """

        @classmethod
        def create(cls):
            """
            Create and setup EC2-instance for remote environment in PyCharm
            """

            # Create security group
            group = aws.connections.ec2.create_security_group('remote-environment', 'Remote environment')
            group.authorize('tcp', 22, 22, '0.0.0.0/0')

            # Create key
            key = aws.helpers.get_or_create_key_pair('remote-environment')

            # Create instance
            instance = aws.helpers.create_instance(
                key.name,
                group.name,
                aws.helpers.find_last_ubuntu_image_id('14.04'),
                name='remote-environment')

            # Connect to the instance via SSH
            key_path = aws.helpers.generate_path_to_key_pair(key.name)
            client = ssh.Client(instance.public_dns_name, key_path=key_path)

            # Configure the instance
            packages = ['python-pip', 'python3-dev', 'build-essential']
            client.apt_install(packages)

            # Show its url
            print('ssh://ubuntu@' + instance.public_dns_name)

        @classmethod
        def delete(cls):
            """
            Remove instance for remote environment
            """

            # Remove instance
            aws.helpers.delete_instance('remote-environment')

            # Remove key
            aws.helpers.delete_key_pair('remote-environment')

            # Remove security group
            aws.connections.ec2.delete_security_group('remote-environment')


import random
import sys
import logging


logger = logging.getLogger(__name__)


def select(message, choices, title=None):
    """
    Show choices and offer to choose one of them

    :type message: str
    :param message: message

    :type choices: list
    :param choices: list of choices

    :type title: callable
    :param title: lambda-function for formatting title of a choice

    :return: choice
    """

    if len(choices) == 0:
        logger.debug('Nothing to choose')

    elif len(choices) == 1:
        logger.debug('Only one choice found')
        return choices[0]

    # Print message and choices
    print('{}:'.format(message))
    for (position, choice) in list(enumerate(choices)):
        print('\t[{position}] {title}'.format(
            position=position,
            title=title(choice) if title else str(choice)
        ))

    # Wait for choice
    while True:
        try:
            selected = int(input('Ваш вариант: '))
            return choices[selected]

        # Incorrect value
        except ValueError:
            continue

        # Index is not from the list
        except IndexError:
            continue

        except KeyboardInterrupt:
            print()
            sys.exit(1)


def confirm():
    """
    Ask for confirmation

    :return: True or False
    """

    code = random.randint(100, 999)
    message = 'Please enter {code} as confirmation code: '.format(code=code)

    # Wait for code been correct
    while True:
        try:
            if code == int(input(message)):
                return True

        # Incorrect value
        except ValueError:
            continue

        except KeyboardInterrupt:
            print()
            sys.exit(1)


def request(message, default=None):
    """
    Request input and return answer

    :type message: str
    :param message: Message

    :type default: str
    :param default: Default answer

    :return: answer or default
    """

    # Add default if need
    message += ' [{}]: '.format(default) if default else ': '
    try:
        answer = input(message)
    except KeyboardInterrupt:
        print()
        sys.exit(1)

    # Return answer or default
    return answer or default


import os
import sys
import time
import logging

import requests
import requests.exceptions
import boto.ec2
import boto.ec2.elb
import boto.ec2.instance
import boto.ec2.autoscale
import boto.ec2.cloudwatch
import boto.ec2.image
import boto.vpc
import boto.exception

from easyaws.libs import interface


logger = logging.getLogger(__name__)


class __Connections:
    """
    Connect to AWS-services via boto and cache these connections
    """

    __cache = dict()
    __services = {
        'ec2': (boto.ec2, 'get_all_regions'),
        'elb': (boto.ec2.elb, 'get_all_load_balancers'),
        'autoscale': (boto.ec2.autoscale, 'get_all_groups'),
        'cloudwatch': (boto.ec2.cloudwatch, 'list_metrics'),
    }

    def __getattr__(self, name):
        """
        Return cached connection to AWS-service or create and cache new connection

        :type name: str
        :param name: name of AWS-service (ec2, elb, autoscale, cloudwatch, etc)

        :return: connection to AWS-service
        """

        if name not in self.__services:
            raise Exception('Unknown service: "{}"'.format(name))

        try:
            if name not in self.__cache:
                # Establish connection
                service = self.__services[name][0]
                region_name = boto.config.get('Boto', 'aws_region_name')
                connection = service.connect_to_region(region_name)

                # Test the connection
                method_name = self.__services[name][1]
                if method_name:
                    getattr(connection, method_name)()

                # Cache the connection
                self.__cache[name] = connection

        except boto.exception.NoAuthHandlerFound:
            logger.critical('Configuration file not found')
            sys.exit(1)

        except boto.exception.EC2ResponseError as e:
            logger.critical(e.message)
            sys.exit(1)

        # Return the connection from cache
        return self.__cache[name]


class helpers:
    """
    Provide helpers for easy manipulation with AWS-services
    """

    __keys_path = os.path.expanduser('~/.ssh/aws')
    __free_tier = {'ec2_instance_type': 't2.micro'}  # Free limits provided by the Free Tier

    @classmethod
    def generate_path_to_key_pair(cls, name):
        """
        Return path to key pair

        :type name: str
        :param name: name of key pair

        :return: path to key pair
        """

        path = os.path.join(cls.__keys_path, '{}.pem'.format(name))
        return path

    @classmethod
    def get_or_create_key_pair(cls, name, save_key_material=True):
        """
        Return existed or new key pair

        :type name: str
        :param name: name of key pair

        :type save_key_material: bool
        :param save_key_material: require to save a key material to a file

        :return: key pair
        """

        logger.debug('Try to get key pair "{}"'.format(name))
        key = connections.ec2.get_key_pair(name)

        if not key:
            logger.debug('Key pair "{}" is not found'.format(name))
            logger.info('Create new key pair "{}"'.format(name))
            key = connections.ec2.create_key_pair(name)

            if save_key_material:
                if not os.path.isdir(cls.__keys_path):
                    logger.debug('Create local directory for key pairs: {}'.format(cls.__keys_path))
                    os.mkdir(cls.__keys_path, 0o0700)

                path = cls.generate_path_to_key_pair(name)
                logger.debug('Save a key material of the key pair "{}" to file: {}'.format(name, path))
                with open(path, 'w') as f:
                    f.write(key.material)
                os.chmod(path, 0o0400)

        return key

    @classmethod
    def delete_key_pair(cls, name):
        """
        Remove key pair from AWS and local filesystem

        :type name: str
        :param name: name of key pair

        :return: result
        """

        key = connections.ec2.get_key_pair(name)
        if not key:
            return False

        logger.info('Remove key pair "{}"'.format(name))
        key.delete()

        path = cls.generate_path_to_key_pair(name)
        if os.path.isfile(path):
            logger.debug('Remove a file of the key pair "{}": {}'.format(name, path))
            os.remove(path)

        return True

    @classmethod
    def get_or_create_security_group(cls, name, public_ports=None):
        """
        Return existed or new security group with open ports

        :type name: str
        :param name: name of security group

        :type public_ports: list
        :param public_ports: list of ports to opening

        :return: security group
        """

        try:
            logger.debug('Try to get security group "{}"'.format(name))
            group = connections.ec2.get_all_security_groups([name])[0]
        except boto.exception.EC2ResponseError as e:
            if e.error_code == 'InvalidGroup.NotFound':
                logger.info('Create security group "{}"'.format(name))
                group = connections.ec2.create_security_group(name, '-')
            else:
                logger.critical(e.message)
                sys.exit(1)

        # Open specified ports
        if public_ports:
            for port in public_ports:
                try:
                    logger.info('Open for all {} port in the security group "{}"'.format(port, name))
                    group.authorize('tcp', port, port, '0.0.0.0/0')
                except boto.exception.EC2ResponseError as e:
                    if e.error_code == 'InvalidPermission.Duplicate':
                        logger.debug('This port is already open')
                        continue
                    else:
                        logger.critical(e.message)
                        sys.exit(1)

        return group

    @classmethod
    def delete_security_group(cls, name):
        """
        Remove security group

        :type name: str
        :param name: name of security group

        :return: result
        """

        try:
            logger.debug('Try to get security group "{}"'.format(name))
            group = connections.ec2.get_all_security_groups([name])[0]
        except boto.exception.EC2ResponseError as e:
            if e.error_code == 'InvalidGroup.NotFound':
                return False
            else:
                logger.critical(e.message)
                sys.exit(1)

        logger.info('Remove security group "{}"'.format(name))
        group.delete()
        return True

    @classmethod
    def create_instance(cls, key_name, group_name, image_id, instance_type=None, name=None):
        """
        Create new EC2-instance and save its name in the tag "name" if specified

        :type key_name: str
        :param key_name: name of key pair

        :type group_name: str
        :param group_name: name of security group

        :type image_id: str
        :param image_id: image id

        :type instance_type: str
        :param instance_type: type of instance

        :type name: str or None
        :param name: name of instance

        :return: result
        """

        if name and cls.get_instance(name):
            logger.critical('The instance "{}" is already exists'.format(name))
            sys.exit(1)

        try:
            logger.info('Create new EC2-instance "{}" based on the image "{}"'.format(name or 'unnamed', image_id))
            reservation = connections.ec2.run_instances(
                image_id,
                key_name=key_name,
                instance_type=instance_type or cls.__free_tier['ec2_instance_type'],
                monitoring_enabled=True,
                security_group_ids=[group_name])
            instance = reservation.instances[0]
        except boto.exception.EC2ResponseError as e:
            if e.error_code == 'InvalidAMIID.NotFound':
                logger.critical('Image with id "{}" is not found'.format(image_id))
            else:
                logger.critical(e.message)
            sys.exit(1)
        if name:
            logger.debug('Save name "{}" of the instance to the tag "name"'.format(name))
            instance.add_tag('name', name)

        # Remember instance id
        instance_id = instance.id

        # Wait instance
        logger.debug('Wait for the instance to be running')
        while True:
            instance = cls.get_instance(instance_id)
            if instance and instance.state == 'running':
                break
            time.sleep(1)
        logger.debug('The instance is running')
        logger.debug('Wait for the instance to be available')
        while True:
            status = cls.get_instance_status(instance_id)
            if status and 'initializing' not in [status.system_status.status, status.instance_status.status]:
                break
            time.sleep(1)
        logger.debug('The instance is available')

        # Reload and return
        return cls.get_instance(instance, reload_from_aws=True)

    @classmethod
    def get_all_instances(cls, instance_ids=None, filters=None):
        """
        Return all alive instances

        :type instance_ids: list
        :param instance_ids: list of instance ids

        :type filters: dict
        :param filters: dictionary of filters

        :return: list of instances
        """

        # Get list of all instances
        reservations = connections.ec2.get_all_instances(instance_ids=instance_ids, filters=filters)
        all_instances = [reservation.instances for reservation in reservations]

        # Filter alive only
        alive_instances = [instance for instance in sum(all_instances, []) if instance.state != 'terminated']
        logger.debug('{alive} alive found from {all} instances'.format(
            alive=len(alive_instances),
            all=len(all_instances)))

        # Return alive instances
        return alive_instances

    @classmethod
    def get_instance(cls, lookup, reload_from_aws=False):
        """
        Return instance

        :type lookup: mixed
        :param lookup: instance, or its id, or its name

        :type reload_from_aws: bool
        :param reload_from_aws: require to reload image object from AWS

        :return: instance or None
        """

        if isinstance(lookup, str):
            # Try to find by name
            response = cls.get_all_instances(filters={'tag:name': lookup})
            if len(response) > 1:
                logger.critical('More than one instance found by name "{}"'.format(lookup))
                sys.exit(1)
            elif len(response) == 1:
                logger.debug('Instance is found by name "{}"'.format(lookup))
                return response[0]

            # Try to find by id
            try:
                response = cls.get_all_instances(instance_ids=[lookup])
                logger.debug('Instance is found by id "{}"'.format(lookup))
                return response[0]
            except boto.exception.EC2ResponseError:
                pass

            logger.debug('Instance is not found by "{}"'.format(lookup))
            return None

        elif isinstance(lookup, boto.ec2.instance.Instance):
            # Reload if need
            return cls.get_instance(lookup.id) if reload_from_aws else lookup

        else:
            raise Exception('Unknown argument type: {}'.format(type(lookup)))

    @classmethod
    def get_instance_status(cls, instance_id):
        """
        Get instance status

        :type instance_id: str
        :param instance_id: instance id

        :return: instance status or None
        """

        response = connections.ec2.get_all_instance_status(instance_ids=[instance_id])
        return response[0] if response else None

    @classmethod
    def delete_instance(cls, instance, wait=True):
        """
        Remove EC2-instance

        :type instance: str or object
        :param instance: instance, or its name, or its id

        :type wait: bool
        :param wait: require to wait for instance to be terminated

        :return: result
        """

        instance = cls.get_instance(instance)
        name = getattr(instance.tags, 'name', None)

        logger.info('Terminate instance "{}"'.format(name or instance.id))
        connections.ec2.terminate_instances(instance_ids=[instance.id])

        if not wait:
            logger.debug('Do not wait for the termination of the instance "{}"'.format(name or instance.id))
            return None

        logger.debug('Wait for the instance "{}" to be terminated'.format(name or instance.id))
        while instance.state != 'terminated':
            time.sleep(1)
            instance.update()

        return True

    @classmethod
    def create_image(cls, name, instance):
        """
        Create new image from the instance

        :type instance: str or object
        :param instance: instance, or its name, or its id

        :return: image
        """

        instance = cls.get_instance(instance)
        instance_name = getattr(instance.tags, 'name', None)

        logger.info('Create image "{}" from the instance "{}"'.format(name, instance_name or instance.id))
        image_id = connections.ec2.create_image(instance.id, name)

        logger.debug('Wait for the image to be saved "{}"'.format(name))
        image = cls.get_image(image_id)
        while image.state == 'pending':
            time.sleep(1)
            image.update()

        return image

    @classmethod
    def get_image(cls, lookup, reload_from_aws=True):
        """
        Return image

        :type lookup: str or object
        :param lookup: image, or its name, or its id

        :type reload_from_aws: bool
        :param reload_from_aws: require to reload image object from AWS

        :return: image or None
        """

        if isinstance(lookup, boto.ec2.image.Image):
            image = cls.get_image(lookup.id) if reload_from_aws else lookup

        elif isinstance(lookup, str) and connections.ec2.get_image(lookup):
            logger.debug('The image is obtained by its id')
            image = connections.ec2.get_image(lookup)

        elif isinstance(lookup, str) and len(connections.ec2.get_all_images(owners=['self'], filters={'name': lookup})):
            logger.debug('The image is obtained by its name')
            image = connections.ec2.get_all_images(owners=['self'], filters={'name': lookup})[0]

        elif isinstance(lookup, str):
            logger.debug('The image "{}" is not found'.format(lookup))
            image = None

        else:
            raise Exception('Unknown argument type: {}'.format(type(lookup)))

        return image

    @classmethod
    def get_all_images(cls, owners=None, filters=None, reverse=False):
        """
        Return all images

        :type filters: dict
        :param filters: filters

        :return: images list
        """

        # Get images
        images = connections.ec2.get_all_images(owners=owners or ['self'], filters=filters)

        # Sort images by name
        images.sort(key=lambda img: img.name, reverse=reverse)

        # Return images
        return images

    @classmethod
    def find_last_ubuntu_image_id(cls, version):
        """
        Find fresh ubuntu image

        :type version: str
        :param version: version, eg. '14.04'

        :return: image id
        """

        cache_id = '/tmp/easy-aws-last-ubuntu-ami'

        logger.debug('Try to find fresh ubuntu {} image via boto'.format(version))
        name = 'ubuntu*hvm*ssd*{version}*'.format(version=version)
        images = cls.get_all_images(
            owners=['099720109477'],  # Canonical id
            filters={'name': name},
            reverse=True)

        if len(images) > 0:
            image_id = images[0].id

        else:
            logger.warning('Can not find any ubuntu image')
            # Try to find last used image
            last_id = open(cache_id).read() if os.path.isfile(cache_id) else None
            # Request image
            image_id = interface.request('Please enter image id', last_id)

        # Save image id to cache and return it
        with open(cache_id, 'w') as f:
            f.write(image_id)
        return image_id

    @classmethod
    def delete_image(cls, lookup):
        """
        Remove image

        :type lookup: str or object
        :param lookup: image, or its name, or its id

        :return: result
        """

        image = cls.get_image(lookup)
        if not image:
            return False

        logger.info('Remove image "{}"'.format(image.name or image.id))
        image.deregister(delete_snapshot=True)
        return True

    @classmethod
    def create_load_balancer(cls, name, security_group, listeners):
        """
        Create load balancer or return already created

        :type name: str
        :param name: name of load balancer

        :type security_group: str
        :param security_group: name of security group

        :type listeners: list
        :param listeners: listening ports [(80, 80, 'http'), (443, 443, 'tcp')]

        :return: load balancer
        """

        logger.info('Create load balancer "{}"'.format(name))

        # Create or setup security group
        public_ports = [listener[0] for listener in listeners]
        security_group = cls.get_or_create_security_group(security_group, public_ports)

        # Get list of available zones in the current region
        zones = [zone.name for zone in connections.ec2.get_all_zones()]

        # Create new load balancer or get already created
        load_balancer = connections.elb.create_load_balancer(
            name,
            zones,
            security_groups=[security_group.id],
            listeners=listeners,
        )
        health_check = boto.ec2.elb.HealthCheck(target='TCP:80')
        load_balancer.configure_health_check(health_check)

        return load_balancer

    @classmethod
    def delete_load_balancer(cls, name):
        """
        Remove load balancer

        :type name: str
        :param name: name of load balancer

        :return: result
        """

        logger.info('Remove load balancer "{}"'.format(name))

        response = connections.elb.get_all_load_balancers(load_balancer_names=[name])
        if not len(response):
            logger.debug('Load balancer "{}" is not found'.format(name))
            return False
        load_balancer = response[0]
        load_balancer.delete()

        logger.debug('Wait for each connected network interface to be removed')
        while len(connections.ec2.get_all_network_interfaces(filters={'description': 'ELB {}'.format(name)})):
            time.sleep(1)

        return True

    @classmethod
    def create_launch_configuration(cls, name, image_id, key_name, security_group, instance_type=None):
        """
        Create launch configuration

        :type name: str
        :param name: name of launch configuration

        :type image_id: str
        :param image_id: image id

        :type key_name: str
        :param key_name: name of key pair

        :type security_group: str
        :param security_group: name of security group

        :type instance_type: str
        :param instance_type: type of instance

        :return: launch configuration
        """

        security_group = cls.get_or_create_security_group(security_group)

        logger.info('Create launch configuration "{}" based on the image "{}"'.format(name, image_id))
        launch_configuration = boto.ec2.autoscale.LaunchConfiguration(
            name=name,
            image_id=image_id,
            key_name=key_name,
            security_groups=[security_group.id],
            instance_type=instance_type or cls.__free_tier['ec2_instance_type'],
            instance_monitoring=True)
        connections.autoscale.create_launch_configuration(launch_configuration)

        return launch_configuration

    @classmethod
    def delete_launch_configuration(cls, name):
        """
        Remove launch configuration

        :type name: str
        :param name: name of launch configuration

        :return: result
        """

        logger.info('Remove launch configuration "{}"'.format(name))
        response = connections.autoscale.get_all_launch_configurations(names=[name])
        if not len(response):
            return False

        connections.autoscale.delete_launch_configuration(name)
        return True

    @classmethod
    def create_auto_scaling_group(cls, name, load_balancer, launch_configuration, min_size=1, max_size=3):
        """
        Create auto scaling group

        :param name: str
        :param name: name of auto scaling group

        :type load_balancer: str
        :param load_balancer: name of load balancer

        :type launch_configuration: str
        :param launch_configuration: name of launch configuration

        :type min_size: int
        :param min_size: minimum number of instance in the group

        :type max_size: int
        :param max_size: maximum number of instance in the group

        :return: auto scaling group
        """

        logger.info('Create auto scaling group "{}"'.format(name))

        # Get list of available zones in the current region
        zones = [zone.name for zone in connections.ec2.get_all_zones()]

        # Create auto scaling group
        auto_scaling_group = boto.ec2.autoscale.AutoScalingGroup(
            group_name=name,
            load_balancers=[load_balancer],
            availability_zones=zones,
            launch_config=launch_configuration,
            min_size=min_size,
            max_size=max_size)
        connections.autoscale.create_auto_scaling_group(auto_scaling_group)

        # Create scaling policies and send them to AWS
        scale_up_policy = boto.ec2.autoscale.ScalingPolicy(
            name='scale_up',
            adjustment_type='ChangeInCapacity',
            as_name=auto_scaling_group.name,
            scaling_adjustment=1,
            cooldown=180)
        scale_down_policy = boto.ec2.autoscale.ScalingPolicy(
            name='scale_down',
            adjustment_type='ChangeInCapacity',
            as_name=auto_scaling_group.name,
            scaling_adjustment=-1,
            cooldown=180)
        connections.autoscale.create_scaling_policy(scale_up_policy)
        connections.autoscale.create_scaling_policy(scale_down_policy)

        # Get them back to obtain policy arns
        scale_up_policy = connections.autoscale.get_all_policies(
            as_group=auto_scaling_group.name,
            policy_names=['scale_up'])[0]
        scale_down_policy = connections.autoscale.get_all_policies(
            as_group=auto_scaling_group.name,
            policy_names=['scale_down'])[0]

        # Create and register alarms
        alarm_dimensions = {"AutoScalingGroupName": auto_scaling_group.name}
        scale_up_alarm = boto.ec2.cloudwatch.MetricAlarm(
            name='scale_up_on_cpu',
            namespace='AWS/EC2',
            metric='CPUUtilization',
            statistic='Average',
            comparison='>',
            threshold=70,
            period=60,
            evaluation_periods=2,
            alarm_actions=[scale_up_policy.policy_arn],  # connecting scaling policy by arn
            dimensions=alarm_dimensions)
        scale_down_alarm = boto.ec2.cloudwatch.MetricAlarm(
            name='scale_down_on_cpu',
            namespace='AWS/EC2',
            metric='CPUUtilization',
            statistic='Average',
            comparison='<',
            threshold=40,
            period=60,
            evaluation_periods=2,
            alarm_actions=[scale_down_policy.policy_arn],  # connecting scaling policy by arn
            dimensions=alarm_dimensions)
        connections.cloudwatch.create_alarm(scale_up_alarm)
        connections.cloudwatch.create_alarm(scale_down_alarm)

        logger.info('Wait for at least one instance been available in the auto scaling group "{}"'.format(name))
        while True:
            auto_scaling_group = cls.get_auto_scaling_group(name)
            in_service = [i.lifecycle_state == 'InService' for i in auto_scaling_group.instances]
            if any(in_service):
                break
            time.sleep(1)

        return auto_scaling_group

    @classmethod
    def get_auto_scaling_group(cls, name):
        """
        Return auto scaling group

        :type name: str
        :param name: name of auto scaling group

        :return: auto scaling group
        """

        response = connections.autoscale.get_all_groups(names=[name])

        if len(response) == 0:
            return None

        return response[0]

    @classmethod
    def delete_auto_scaling_group(cls, name):
        """
        Remove auto scaling group

        :type name: str
        :param name: name of auto scaling group

        :return: result
        """

        logger.info('Remove the auto scaling group "{}"'.format(name))
        auto_scaling_group = cls.get_auto_scaling_group(name)
        if not auto_scaling_group:
            return False

        # Remove
        auto_scaling_group.delete(force_delete=True)

        logger.debug('Wait for the auto scaling group to be deleted')
        while cls.get_auto_scaling_group(name):
            time.sleep(1)
        return True

    @classmethod
    def wait_host(cls, dns_name):
        """
        Wait for host availability

        :type dns_name: str
        :param dns_name: url, eg. "example.com"
        """

        url = 'http://{}/'.format(dns_name)
        logger.info('Wait for host to be available by url: {}'.format(url))
        while True:
            try:
                # Send HEAD-request to the host
                response = requests.get(url, timeout=3)
                if response.status_code == 200:
                    break
                time.sleep(1)
            except requests.exceptions.ConnectionError:
                time.sleep(1)
            except KeyboardInterrupt:
                sys.exit(1)

        logger.debug('Host is available')


def get_regions():
    """
    Return list of available regions for using services

    :return: list
    """

    # Get regions for each service
    ec2_regions = [region.name for region in boto.ec2.regions()]
    foo_regions = ec2_regions[:]  # Example of regions of other service

    # Find intersection
    regions = set(ec2_regions).intersection(foo_regions)

    # Return sorted list
    return sorted(regions)

# Return connections to the AWS-services
connections = __Connections()


import os


def numCPUs():
    if not hasattr(os, "sysconf"):
        raise RuntimeError("No sysconf detected.")
    return os.sysconf("SC_NPROCESSORS_ONLN")


bind = "127.0.0.1:8000"
workers = numCPUs() * 2 + 1


#!/usr/bin/env python3
import inspect
import logging

import argparse

from easyaws.commands import Commands


def generate_parsers(parser, klass):
    """
    Generate subparsers based on the class that describes hierarchy of commands

    :type parser: object
    :param parser: parent parser

    :type klass: class
    :param klass: class that describes hierarchy of commands
    """

    # Get list of commands in the class
    commands = [
        (command.lower(), getattr(klass, command))  # (имя команды, объект команды)
        for command in dir(klass) if not command.startswith('__')
    ]

    # Add subparsers
    subparsers = parser.add_subparsers()
    for name, command in commands:
        # Get command description from properties __docstring__ or __doc__
        doc = inspect.getdoc(command)
        doc = doc[0].lower() + doc[1:] if doc else ''
        subparser = subparsers.add_parser(name, help=doc)

        # Method is detected, i.e single command
        if inspect.ismethod(command):
            subparser.set_defaults(command=command)

        # Class is detected, i.e. group of commands
        elif inspect.isclass(command):
            generate_parsers(subparser, command)


def main():
    # Create main parser
    parser = argparse.ArgumentParser(description='Tool for easy deploying django-projects to Amazon AWS')
    parser.add_argument('-d', '--debug', action='store_true', help='show debug log messages')
    generate_parsers(parser, Commands)

    # Parse arguments
    args = parser.parse_args()

    # Setup third-party loggers
    logging.getLogger('boto').setLevel(logging.CRITICAL)
    logging.getLogger('requests').setLevel(logging.CRITICAL)
    logging.getLogger('paramiko').setLevel(logging.CRITICAL)

    # Create logger
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    # Create console handler
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG if args.debug else logging.INFO)
    console_formatter = logging.Formatter('{message}', style='{')
    console.setFormatter(console_formatter)

    # Create file handler
    file = logging.FileHandler('/tmp/{}.log'.format(__name__))
    file.setLevel(logging.DEBUG)
    file_formatter = logging.Formatter('{asctime} {levelname: <8} {name}: {message}', style='{')
    file.setFormatter(file_formatter)

    # Add handlers to logger
    logger.addHandler(console)
    logger.addHandler(file)

    # Execute command
    if hasattr(args, 'command'):
        print(inspect.getdoc(args.command))
        args.command()

    # Or show main help
    else:
        parser.print_help()


import sys
import socket
import tempfile
import string
import time
import logging

import pkg_resources
import paramiko
import paramiko.ssh_exception


logger = logging.getLogger(__name__)


class ClientError(BaseException):
    """
    Exception for this module
    """

    def __init__(self, message):
        """
        :type message: str
        :param message: error message
        """

        self.message = message
        logger.critical(self.message)


class Client:
    """
    Client for tasks via SSH
    """

    def __init__(self, url, username='ubuntu', key_path=None, replacements=None, timeout=30):
        """
        Connect to remote host

        :type url: str
        :param url: remote host

        :type username: str
        :param username: username

        :type key_path: str
        :param key_path: path to a file of the key

        :type replacements: dict
        :param replacements: dictionary to replacement in commands and files

        :type timeout: int
        :param timeout: timeout in seconds

        :return: API
        """

        self.__replacements = replacements or dict()

        logger.info('Connect to "{username}@{url}" via SSH'.format(username=username, url=url))

        logger.debug('Wait for 22 port to be opened')
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(0.2)
        while sock.connect_ex((url, 22)) is not 0:
            time.sleep(1)

        logger.debug('Establish connection')
        self.__client = paramiko.SSHClient()
        self.__client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            self.__client.connect(url, username=username, timeout=timeout, key_filename=key_path)
        except socket.timeout:
            raise ClientError('Stopped on timeout in {} seconds'.format(timeout))
        except paramiko.ssh_exception.AuthenticationException:
            raise ClientError('Authentication error')

    def run_command(self, command, timeout=30):
        """
        Execute command on the remote host

        :type command: str
        :param command: command

        :type timeout: int
        :param timeout: timeout in seconds
        """

        # Make replacement
        command = command.format(**self.__replacements)
        logger.debug('$ {}'.format(command))

        try:
            # Send command without blocking and redirect stderr to stdout and wait for completion
            stdin, stdout, stderr = self.__client.exec_command(command + ' 2>&1', timeout=timeout)
            time.sleep(0.1)
            while not stdout.channel.exit_status_ready():
                time.sleep(1)
        except socket.timeout:
            logger.critical('Stopped on timeout in {} seconds'.format(timeout))
            sys.exit(1)

        if stdout.channel.exit_status != 0:
            print(''.join(stdout.readlines()))
            raise ClientError('Command completed with error')

    def run_commands(self, commands):
        """
        Executes list of commands

        :type commands: list
        :param commands: list of commands
        """

        logger.info('Execute {} commands on the remote host'.format(len(commands)))
        for command in commands:
            self.run_command(command)

    def put_files(self, files):
        """
        Put files from local filesystem to the remote host, produces automatic variable replacement.

        Rules for replacements:
            $PROJECT_PATH
                or
            ${PROJECT_PATH}

        Example:
            files = [
                ('templates/supervisord.conf', '/tmp/my.supervisord.conf'),
                ('templates/gunicorn.conf.py', '{PROJECT_PATH}/gunicorn.conf.py'),
                ('/absolute/path/file.py', '/tmp/file.py'),
            ]
            ssh.put_files(files)

        :type files: list
        :param files: list, eg. [('templates/gunicorn.conf.py', '{PROJECT_PATH}/gunicorn.conf.py'), ...]
        """

        logger.info('Put {} files to the remote host'.format(len(files)))
        sftp = self.__client.open_sftp()
        for (source_path, remote_path) in files:
            # Convert relative path of the source file to absolute path
            if not str(source_path).startswith('/'):
                source_path = pkg_resources.resource_filename('easyaws', source_path)

            logger.debug('{} -> {}'.format(source_path, remote_path))
            try:
                # Try to read content of the source file
                source_content = open(source_path).read()
            except UnicodeDecodeError:
                # Can't read content (eg. it is a zipped tar-archive)
                pass
            else:
                # Create new content based on the template after replacement
                template = string.Template(source_content)
                new_content = template.safe_substitute(**self.__replacements)
                if source_content != new_content:
                    # Create temp file
                    tmp = tempfile.NamedTemporaryFile(delete=False)
                    tmp.write(new_content.encode('utf-8') + b'\n')
                    tmp.close()
                    # Change source path to the created temp file
                    source_path = tmp.name

            # Put the source file to the remote host
            try:
                sftp.put(source_path, remote_path)
            except FileNotFoundError:
                raise ClientError('Error in the remote path')

        # Close connection
        sftp.close()

    def apt_install(self, packages):
        """
        Install packages via apt-get

        :type packages: list
        :param packages: list of packages
        """

        commands = ['sudo apt-get update',
                    'sudo apt-get install {} -y'.format(' '.join(packages))]

        for command in commands:
            try:
                self.run_command(command)
            except ClientError:
                # Try again on error
                self.run_command(command)

    def __del__(self):
        self.__client.close()


import os
import inspect

from setuptools import setup
from setuptools import find_packages


__location__ = os.path.join(os.getcwd(), os.path.dirname(inspect.getfile(inspect.currentframe())))
setup(
    name='easy-aws',
    description='Deploying django-projects to Amazon AWS',
    long_description=open(os.path.join(__location__, 'README.rst')).read(),
    version='1.0.3',

    # The project's main homepage
    url='https://bitbucket.org/dlis/easy-aws',

    # Author details
    author='dlis',
    author_email='lisovsky@gmail.com',

    # Project license
    license='MIT',

    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # Development Status
        'Development Status :: 5 - Production/Stable',

        # Classification
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Topic :: Utilities',

        # License
        'License :: OSI Approved :: MIT License',

        # Python versions
        'Programming Language :: Python :: 3.4',
    ],

    # Relating keywords
    keywords='deploy django amazon aws ec2 elb pycharm',

    # Packages and its data
    packages=find_packages(),
    package_data={'easyaws': ['templates/*']},

    # Requirements
    install_requires=[
        'argparse',
        'paramiko',
        'boto',
        'requests',
    ],

    # Entry points
    entry_points={
        'console_scripts': [
            'easy-aws=easyaws:main',
        ],
    },
)